from .agent import SingleRoomAgent
from meshagent.api.chan import Chan
from meshagent.api import RoomMessage, RoomClient
from meshagent.agents import AgentChatContext
from meshagent.tools import RemoteToolkit, Tool, Toolkit, make_toolkits, ToolkitBuilder
from .adapter import LLMAdapter, ToolResponseAdapter
import asyncio
from typing import Optional
import json
from meshagent.tools import ToolContext
import logging

from pathlib import Path
from meshagent.agents.skills import to_prompt

logger = logging.getLogger("worker")


class SubmitWork(Tool):
    def __init__(self, *, agent: "Worker", queue: str):
        self.queue = queue
        self.agent = agent
        super().__init__(
            name=f"queue_{agent.name}_task",
            title=f"Queue {agent.title} Task",
            description=f"Queues a new task to the worker -- {agent.description}",
            input_schema={
                "type": "object",
                "required": ["prompt"],
                "additionalProperties": False,
                "properties": {
                    "prompt": {
                        "type": "string",
                    },
                },
            },
        )

    async def execute(self, context: ToolContext, *, prompt: str):
        await context.room.queues.send(
            name=self.queue,
            message={
                "prompt": prompt,
            },
            create=True,
        )
        return None


class Worker(SingleRoomAgent):
    def __init__(
        self,
        *,
        queue: str,
        name=None,
        title=None,
        description=None,
        requires=None,
        llm_adapter: LLMAdapter,
        tool_adapter: Optional[ToolResponseAdapter] = None,
        toolkits: Optional[list[Toolkit]] = None,
        rules: Optional[list[str]] = None,
        toolkit_name: Optional[str] = None,
        skill_dirs: Optional[list[str]] = None,
        supports_context: bool = True,
    ):
        super().__init__(
            name=name,
            title=title,
            description=description,
            requires=requires,
        )

        self._skill_dirs = skill_dirs

        self._queue = queue

        if toolkits is None:
            toolkits = []

        self._llm_adapter = llm_adapter
        self._tool_adapter = tool_adapter

        self._message_channel = Chan[RoomMessage]()

        self._room: RoomClient | None = None
        self._toolkits = toolkits

        if rules is None:
            rules = []

        self._rules = rules
        self._done = False

        if toolkit_name is not None:
            logger.info(f"worker will start toolkit {toolkit_name}")
            self._worker_toolkit = RemoteToolkit(
                name=toolkit_name,
                tools=[
                    SubmitWork(queue=self._queue, agent=self),
                ],
            )
        else:
            self._worker_toolkit = None

        self.supports_context = supports_context

    async def start(self, *, room: RoomClient):
        self._done = False

        if self._worker_toolkit is not None:
            await self._worker_toolkit.start(room=room)

        await super().start(room=room)

        self._main_task = asyncio.create_task(self.run(room=room))

    async def stop(self):
        self._done = True

        await asyncio.gather(self._main_task)

        if self._worker_toolkit is not None:
            await self._worker_toolkit.stop()

        await super().stop()

    async def get_rules(self):
        rules = [*self._rules]

        if self._skill_dirs is not None and len(self._skill_dirs) > 0:
            rules.append(
                "You have access to to following skills which follow the agentskills spec:"
            )
            rules.append(await to_prompt([*(Path(p) for p in self._skill_dirs)]))
            rules.append(
                "Use the shell tool to find out more about skills and execute them when they are required"
            )

        return rules

    def get_prompt_for_message(self, *, message: dict) -> str:
        prompt = message.get("prompt")
        if prompt is None:
            logger.warning(
                "prompt property not found on worker message, inserting whole message into context"
            )
            prompt = json.dumps(message)

        return prompt

    async def append_message_context(
        self, *, message: dict, chat_context: AgentChatContext
    ):
        if self.supports_context:
            caller_context_json = message.get("caller_context")
            if caller_context_json is not None:
                caller_context = AgentChatContext.from_json(caller_context_json)

                chat_context.messages.extend(caller_context.messages)
                chat_context.previous_response_id = caller_context.previous_response_id

        prompt = self.get_prompt_for_message(message=message)

        chat_context.append_user_message(message=prompt)

    async def process_message(
        self,
        *,
        chat_context: AgentChatContext,
        message: dict,
        toolkits: list[Toolkit],
    ):
        await self.append_message_context(message=message, chat_context=chat_context)

        return await self._llm_adapter.next(
            context=chat_context,
            room=self.room,
            toolkits=toolkits,
            tool_adapter=self._tool_adapter,
        )

    def get_toolkit_builders(self) -> list[ToolkitBuilder]:
        return []

    async def get_message_toolkits(self, *, message: dict) -> list[Toolkit]:
        toolkits = await self.get_required_toolkits(
            context=ToolContext(
                room=self.room,
                caller=self.room.local_participant,
                on_behalf_of=None,
            )
        )

        tool_providers = [*self.get_toolkit_builders()]

        model = message.get("model", self._llm_adapter.default_model())

        message_tools = message.get("tools")

        if message_tools is not None and len(message_tools) > 0:
            toolkits.extend(
                await make_toolkits(
                    room=self.room,
                    model=model,
                    providers=tool_providers,
                    tools=message_tools,
                )
            )
        return [*self._toolkits, *toolkits]

    def prepare_chat_context(self, *, chat_context: AgentChatContext):
        pass

    async def run(self, *, room: RoomClient):
        backoff = 0
        while not self._done:
            try:
                message = await room.queues.receive(
                    name=self._queue, create=True, wait=True
                )

                backoff = 0
                if message is not None:
                    logger.info("received message on worker queue")
                    try:
                        chat_context = await self.init_chat_context()

                        chat_context.replace_rules(
                            rules=[
                                *await self.get_rules(),
                            ]
                        )

                        toolkits = await self.get_message_toolkits(message=message)

                        self.prepare_chat_context(chat_context=chat_context)

                        await self.process_message(
                            chat_context=chat_context,
                            message=message,
                            toolkits=toolkits,
                        )

                    except Exception as e:
                        logger.error(f"Failed to process: {e}\n{message}", exc_info=e)

            except Exception as e:
                logger.error(
                    f"Worker error while receiving: {e}, will retry", exc_info=e
                )

                await asyncio.sleep(0.1 * pow(2, backoff))
                backoff = backoff + 1



from meshagent.api.schema import (
    ElementProperty,
    ValueProperty,
    MeshSchema,
    ChildProperty,
    ElementType,
)
from meshagent.api.schema_document import Document

schema_schema = MeshSchema(
    root_tag_name="schema",
    elements=[
        ElementType(
            tag_name="schema",
            description="defines an xml document representing a json schema",
            properties=[
                ValueProperty(
                    name="root_tag_name",
                    description="the root tag name in this document (must match one of the elements defined in the schema)",
                    type="string",
                ),
                ChildProperty(
                    name="elements",
                    description="the configuration of this sandbox",
                    child_tag_names=["element"],
                ),
            ],
        ),
        ElementType(
            tag_name="element",
            description="an element inside the document",
            properties=[
                ValueProperty(
                    name="tag_name",
                    description="the tag name used by this element. Element names are case-sensitive. Element names cannot contain spaces. Element names must start with a letter or underscore. Element names cannot start with the letters xml (or XML, or Xml, etc) Element names can contain letters, digits, hyphens, underscores, and periods.",
                    type="string",
                ),
                ChildProperty(
                    name="schema",
                    description="the attributes of this node",
                    child_tag_names=["attribute", "child_elements"],
                ),
                ValueProperty(name="description", type="string"),
            ],
        ),
        ElementType(
            tag_name="attribute",
            properties=[
                ValueProperty(
                    name="name",
                    description="the attribute name. Attribute names are case-sensitive. Attribute names cannot contain spaces. Attribute names must start with a letter or underscore. Attribute names cannot start with the letters xml (or XML, or Xml, etc) Attribute names can contain letters, digits, hyphens, underscores, and periods.",
                    type="string",
                ),
                ValueProperty(
                    name="type",
                    description="the type of value stored by this attribute",
                    type="string",
                    enum=["string", "boolean", "number"],
                ),
                ValueProperty(name="description", type="string"),
            ],
        ),
        ElementType(
            tag_name="child_elements",
            properties=[
                ValueProperty(
                    name="name",
                    description="the name of the property to store the children in. Propterty names cannot contain spaces. Propterty names must start with a letter or underscore. Propterty names cannot start with the letters xml (or XML, or Xml, etc) Propterty names can contain letters, digits, hyphens, and underscores.",
                    type="string",
                ),
                ValueProperty(name="description", type="string"),
                ChildProperty(
                    name="children",
                    description="the tag name of an element. The element must have a corresponding element definition under the schema elements",
                    child_tag_names=["element_reference"],
                ),
            ],
        ),
        ElementType(
            tag_name="element_reference",
            description="an allowed child element for an element, should only used when a schema can contain multiple items of a specific type",
            properties=[
                ValueProperty(
                    name="tag_name",
                    description="the tag name of an element. The element must have a corresponding element definition under the schema elements",
                    type="string",
                ),
            ],
        ),
    ],
)


def schema_from_schema_document(document: Document):
    root_tag_name = document.root["root_tag_name"]

    elements = list[ElementType]()

    for element in document.root.get_children():
        properties = list[ElementProperty]()

        for prop in element.get_children():
            name = prop["name"]
            description = prop["description"]
            type = prop["type"]

            if prop.tag_name == "attribute":
                properties.append(
                    ValueProperty(name=name, type=type, description=description)
                )
            elif prop.tag_name == "child_elements":
                child_tag_names = []

                for child in prop.get_children():
                    child_tag_names.append(child["tag_name"])

                properties.append(
                    ChildProperty(name=name, child_tag_names=child_tag_names)
                )

        et = ElementType(
            tag_name=element["tag_name"],
            properties=properties,
            description=element["description"],
        )
        elements.append(et)

    doc = MeshSchema(root_tag_name=root_tag_name, elements=elements)
    return doc



from meshagent.agents.task_runner import TaskContext
from meshagent.api.room_server_client import RoomClient
from meshagent.api.schema_document import Document
from meshagent.api import Requirement

# from meshagent.tools.document_tools import DocumentToolkit
from typing import Optional
import asyncio
import logging
from meshagent.agents.agent import TaskRunner

logger = logging.getLogger("writer_agent")


class WriterContext:
    def __init__(
        self,
        room: RoomClient,
        document: Document,
        call_context: TaskContext,
        path: str,
    ):
        self._room = room
        self._document = document
        self._call_context = call_context
        self._path = path

    @property
    def room(self) -> RoomClient:
        return self._room

    @property
    def path(self) -> str:
        return self._path

    @property
    def document(self) -> Document:
        return self._document

    @property
    def call_context(self) -> TaskContext:
        return self._call_context


# writes to a path, the document must already exist, optionally create the document if it does not exist
class Writer(TaskRunner):
    def __init__(
        self,
        name: str,
        input_schema: dict,
        output_schema: dict,
        create: bool = True,
        title: Optional[str] = None,
        description: Optional[str] = None,
        requires: Optional[list[Requirement]] = None,
        supports_tools: Optional[bool] = None,
    ):
        super().__init__(
            name=name,
            input_schema=input_schema,
            output_schema=output_schema,
            description=description,
            title=title,
            requires=requires,
            supports_tools=supports_tools,
        )

        self._create = create

        if "path" not in input_schema["properties"]:
            raise Exception("input schema must include a 'path' property")

    def pop_path(self, arguments: dict):
        return arguments["path"]

    async def write(self, writer_context: WriterContext, arguments: dict) -> dict:
        pass

    async def ask(self, *, context: TaskContext, arguments: dict):
        path = arguments["path"]

        client = context.room
        document = await client.sync.open(path=path, create=self._create)

        writer_context = WriterContext(
            room=client, document=document, call_context=context, path=path
        )

        try:
            return await self.write(writer_context, arguments)
        finally:
            # TODO: need a way to wait for changes to be synchronized
            await asyncio.sleep(1)
            await client.sync.close(path=path)
